package java.security;

import java.lang.reflect.Constructor;
import java.lang.ref.WeakReference;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.lang.ref.Reference;
import java.util.List;
import java.util.Objects;
import java.util.LinkedHashSet;
import java.util.Locale;
import java.util.LinkedHashMap;
import java.util.Iterator;
import java.util.HashMap;
import java.io.ObjectInputStream;
import java.util.Enumeration;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.function.BiFunction;
import java.util.Collection;
import java.util.Collections;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.ConcurrentHashMap;
import java.util.Map;
import java.util.Set;
import sun.security.util.Debug;
import java.util.Properties;

public abstract class Provider extends Properties
{
    static final long serialVersionUID = -4298000515446427739L;
    private static final Debug debug;
    private String name;
    private String info;
    private double version;
    private transient Set<Map.Entry<Object, Object>> entrySet;
    private transient int entrySetCallCount;
    private transient boolean initialized;
    private transient boolean legacyChanged;
    private transient volatile boolean servicesChanged;
    private transient Map<String, String> legacyStrings;
    private transient Map<Provider.ServiceKey, Provider.Service> serviceMap;
    private transient Set<String> prngAlgos;
    private transient Map<Provider.ServiceKey, Provider.Service> legacyMap;
    private transient Set<Provider.Service> serviceSet;
    private static final String ALIAS_PREFIX = "Alg.Alias.";
    private static final String ALIAS_PREFIX_LOWER = "alg.alias.";
    private static final int ALIAS_LENGTH;
    private static volatile Provider.ServiceKey previousKey;
    private static final Map<String, Provider.EngineDescription> knownEngines;
    
    protected Provider(String name, double version, String info) {
        this.entrySet = null;
        this.entrySetCallCount = 0;
        this.name = name;
        this.version = version;
        this.info = info;
        this.serviceMap = new ConcurrentHashMap();
        this.putId();
        this.initialized = true;
    }
    
    public String getName() {
        return this.name;
    }
    
    public double getVersion() {
        return this.version;
    }
    
    public String getInfo() {
        return this.info;
    }
    
    @Override
    public String toString() {
        return this.name + " version " + this.version;
    }
    
    @Override
    public synchronized void clear() {
        this.check("clearProviderProperties." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Remove " + this.name + " provider properties");
        }
        this.implClear();
    }
    
    @Override
    public synchronized void load(InputStream inStream) throws IOException {
        this.check("putProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Load " + this.name + " provider properties");
        }
        Properties properties = new Properties();
        properties.load(inStream);
        this.implPutAll(properties);
    }
    
    @Override
    public synchronized void putAll(Map<?, ?> map) {
        this.check("putProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Put all " + this.name + " provider properties");
        }
        this.implPutAll(map);
    }
    
    @Override
    public synchronized Set<Map.Entry<Object, Object>> entrySet() {
        this.checkInitialized();
        if (this.entrySet == null) {
            if (this.entrySetCallCount++ == 0) {
                this.entrySet = Collections.unmodifiableMap(this).entrySet();
            }
            else {
                return super.entrySet();
            }
        }
        if (this.entrySetCallCount != 2) {
            throw new RuntimeException("Internal error.");
        }
        return this.entrySet;
    }
    
    @Override
    public Set<Object> keySet() {
        this.checkInitialized();
        return Collections.unmodifiableSet(super.keySet());
    }
    
    @Override
    public Collection<Object> values() {
        this.checkInitialized();
        return Collections.unmodifiableCollection(super.values());
    }
    
    @Override
    public synchronized Object put(Object obj, Object obj2) {
        this.check("putProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Set " + this.name + " provider property [" + obj + "/" + obj2 + "]");
        }
        return this.implPut(obj, obj2);
    }
    
    @Override
    public synchronized Object putIfAbsent(Object obj, Object obj2) {
        this.check("putProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Set " + this.name + " provider property [" + obj + "/" + obj2 + "]");
        }
        return this.implPutIfAbsent(obj, obj2);
    }
    
    @Override
    public synchronized Object remove(Object obj) {
        this.check("removeProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Remove " + this.name + " provider property " + obj);
        }
        return this.implRemove(obj);
    }
    
    @Override
    public synchronized boolean remove(Object obj, Object o) {
        this.check("removeProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Remove " + this.name + " provider property " + obj);
        }
        return this.implRemove(obj, o);
    }
    
    @Override
    public synchronized boolean replace(Object obj, Object o, Object o2) {
        this.check("putProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Replace " + this.name + " provider property " + obj);
        }
        return this.implReplace(obj, o, o2);
    }
    
    @Override
    public synchronized Object replace(Object obj, Object o) {
        this.check("putProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Replace " + this.name + " provider property " + obj);
        }
        return this.implReplace(obj, o);
    }
    
    @Override
    public synchronized void replaceAll(BiFunction<? super Object, ? super Object, ?> biFunction) {
        this.check("putProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("ReplaceAll " + this.name + " provider property ");
        }
        this.implReplaceAll(biFunction);
    }
    
    @Override
    public synchronized Object compute(Object obj, BiFunction<? super Object, ? super Object, ?> biFunction) {
        this.check("putProviderProperty." + this.name);
        this.check("removeProviderProperty" + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Compute " + this.name + " provider property " + obj);
        }
        return this.implCompute(obj, biFunction);
    }
    
    @Override
    public synchronized Object computeIfAbsent(Object obj, Function<? super Object, ?> function) {
        this.check("putProviderProperty." + this.name);
        this.check("removeProviderProperty" + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("ComputeIfAbsent " + this.name + " provider property " + obj);
        }
        return this.implComputeIfAbsent(obj, function);
    }
    
    @Override
    public synchronized Object computeIfPresent(Object obj, BiFunction<? super Object, ? super Object, ?> biFunction) {
        this.check("putProviderProperty." + this.name);
        this.check("removeProviderProperty" + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("ComputeIfPresent " + this.name + " provider property " + obj);
        }
        return this.implComputeIfPresent(obj, biFunction);
    }
    
    @Override
    public synchronized Object merge(Object obj, Object o, BiFunction<? super Object, ? super Object, ?> biFunction) {
        this.check("putProviderProperty." + this.name);
        this.check("removeProviderProperty" + this.name);
        if (Provider.debug != null) {
            Provider.debug.println("Merge " + this.name + " provider property " + obj);
        }
        return this.implMerge(obj, o, biFunction);
    }
    
    @Override
    public Object get(Object key) {
        this.checkInitialized();
        return super.get(key);
    }
    
    @Override
    public synchronized Object getOrDefault(Object key, Object defaultValue) {
        this.checkInitialized();
        return super.getOrDefault(key, defaultValue);
    }
    
    @Override
    public synchronized void forEach(BiConsumer<? super Object, ? super Object> action) {
        this.checkInitialized();
        super.forEach(action);
    }
    
    @Override
    public Enumeration<Object> keys() {
        this.checkInitialized();
        return super.keys();
    }
    
    @Override
    public Enumeration<Object> elements() {
        this.checkInitialized();
        return super.elements();
    }
    
    @Override
    public String getProperty(String key) {
        this.checkInitialized();
        return super.getProperty(key);
    }
    
    private void checkInitialized() {
        if (!this.initialized) {
            throw new IllegalStateException();
        }
        return;
    }
    
    private void check(String target) {
        this.checkInitialized();
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkSecurityAccess(target);
        }
    }
    
    private void putId() {
        super.put("Provider.id name", this.name);
        super.put("Provider.id version", String.valueOf(this.version));
        super.put("Provider.id info", this.info);
        super.put("Provider.id className", this.getClass().getName());
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        HashMap<Object, Object> hashMap = new HashMap();
        for (Map.Entry entry : super.entrySet()) {
            hashMap.put(entry.getKey(), entry.getValue());
        }
        this.defaults = null;
        objectInputStream.defaultReadObject();
        this.serviceMap = new ConcurrentHashMap();
        this.implClear();
        this.initialized = true;
        this.putAll(hashMap);
    }
    
    private boolean checkLegacy(Object o) {
        if (((String)o).startsWith("Provider.")) {
            return false;
        }
        this.legacyChanged = true;
        if (this.legacyStrings == null) {
            this.legacyStrings = new LinkedHashMap();
        }
        return true;
    }
    
    private void implPutAll(Map<?, ?> map) {
        for (Map.Entry entry : map.entrySet()) {
            this.implPut(entry.getKey(), entry.getValue());
        }
    }
    
    private Object implRemove(Object key) {
        if ((key instanceof String)) {
            if (!this.checkLegacy(key)) {
                return null;
            }
            this.legacyStrings.remove((Object)key);
        }
        return super.remove(key);
    }
    
    private boolean implRemove(Object key, Object value) {
        if (key instanceof String && value instanceof String) {
            if (!this.checkLegacy(key)) {
                return false;
            }
            this.legacyStrings.remove((Object)key, (Object)value);
        }
        return super.remove(key, value);
    }
    
    private boolean implReplace(Object key, Object oldValue, Object newValue) {
        if (key instanceof String && oldValue instanceof String && newValue instanceof String) {
            if (!this.checkLegacy(key)) {
                return false;
            }
            this.legacyStrings.replace((String)key, (String)oldValue, (String)newValue);
        }
        return super.replace(key, oldValue, newValue);
    }
    
    private Object implReplace(Object key, Object value) {
        if (key instanceof String && value instanceof String) {
            if (!this.checkLegacy(key)) {
                return null;
            }
            this.legacyStrings.replace((String)key, (String)value);
        }
        return super.replace(key, value);
    }
    
    private void implReplaceAll(BiFunction<? super Object, ? super Object, ?> biFunction) {
        this.legacyChanged = true;
        if (this.legacyStrings == null) {
            this.legacyStrings = new LinkedHashMap();
        }
        else {
            this.legacyStrings.replaceAll(biFunction);
        }
        super.replaceAll(biFunction);
    }
    
    private Object implMerge(Object key, Object value, BiFunction<? super Object, ? super Object, ?> biFunction) {
        if (key instanceof String && value instanceof String) {
            if (!this.checkLegacy(key)) {
                return null;
            }
            this.legacyStrings.merge((String)key, (String)value, biFunction);
        }
        return super.merge(key, value, biFunction);
    }
    
    private Object implCompute(Object key, BiFunction<? super Object, ? super Object, ?> biFunction) {
        if ((key instanceof String)) {
            if (!this.checkLegacy(key)) {
                return null;
            }
            this.legacyStrings.compute((String)key, biFunction);
        }
        return super.compute(key, biFunction);
    }
    
    private Object implComputeIfAbsent(Object key, Function<? super Object, ?> function) {
        if ((key instanceof String)) {
            if (!this.checkLegacy(key)) {
                return null;
            }
            this.legacyStrings.computeIfAbsent((String)key, function);
        }
        return super.computeIfAbsent(key, function);
    }
    
    private Object implComputeIfPresent(Object key, BiFunction<? super Object, ? super Object, ?> biFunction) {
        if ((key instanceof String)) {
            if (!this.checkLegacy(key)) {
                return null;
            }
            this.legacyStrings.computeIfPresent((String)key, biFunction);
        }
        return super.computeIfPresent(key, biFunction);
    }
    
    private Object implPut(Object key, Object value) {
        if (key instanceof String && value instanceof String) {
            if (!this.checkLegacy(key)) {
                return null;
            }
            this.legacyStrings.put((String)key, (String)value);
        }
        return super.put(key, value);
    }
    
    private Object implPutIfAbsent(Object key, Object value) {
        if (key instanceof String && value instanceof String) {
            if (!this.checkLegacy(key)) {
                return null;
            }
            this.legacyStrings.putIfAbsent((String)key, (String)value);
        }
        return super.putIfAbsent(key, value);
    }
    
    private void implClear() {
        if (this.legacyStrings != null) {
            this.legacyStrings.clear();
        }
        if (this.legacyMap != null) {
            this.legacyMap.clear();
        }
        this.serviceMap.clear();
        this.legacyChanged = false;
        this.servicesChanged = false;
        this.serviceSet = null;
        this.prngAlgos = null;
        super.clear();
        this.putId();
    }
    
    private void ensureLegacyParsed() {
        if (!this.legacyChanged || this.legacyStrings == null) {
            return;
        }
        this.serviceSet = null;
        if (this.legacyMap == null) {
            this.legacyMap = new ConcurrentHashMap();
        }
        else {
            this.legacyMap.clear();
        }
        for (Map.Entry entry : this.legacyStrings.entrySet()) {
            this.parseLegacyPut((String)entry.getKey(), (String)entry.getValue());
        }
        removeInvalidServices(this.legacyMap, this.serviceMap);
        this.legacyChanged = false;
        return;
    }
    
    private static void removeInvalidServices(Map<Provider.ServiceKey, Provider.Service> map, Map<Provider.ServiceKey, Provider.Service> map2) {
        Iterator<Object> iterator = map.entrySet().iterator();
        while (iterator.hasNext()) {
            Map.Entry<Provider.Service, Object> entry = (Map.Entry)iterator.next();
            Provider.Service service = (Provider.Service)entry.getValue();
            if (!Provider.Service.access$000(service)) {
                Provider.ServiceKey obj = (Provider.ServiceKey)entry.getKey();
                if (!obj.algorithm.equals(service.algorithm) && service.className == null) {
                    Provider.Service value = (Provider.Service)map2.get(new Provider.ServiceKey(service.type, service.algorithm, true));
                    if (value != null) {
                        entry.setValue(value);
                        continue;
                    }
                }
                if (Provider.debug != null) {
                    Provider.debug.println("Remove invalid legacy mapping under " + obj);
                }
                iterator.remove();
            }
            else {
                continue;
            }
        }
    }
    
    private static String[] getTypeAndAlgorithm(String str) {
        int index = str.indexOf(46);
        if (index < 1) {
            if (Provider.debug != null) {
                Provider.debug.println("Ignoring invalid entry in provider: " + str);
            }
            return null;
        }
        else {
            return new String[] { str.substring(0, index), str.substring(index + 1) };
        }
    }
    
    private void parseLegacyPut(String s, String s2) {
        if (s.toLowerCase(Locale.ENGLISH).startsWith("alg.alias.")) {
            String[] typeAndAlgorithm2 = getTypeAndAlgorithm(s.substring(Provider.ALIAS_LENGTH));
            if (typeAndAlgorithm2 == null) {
                return;
            }
            String engineName3 = getEngineName(typeAndAlgorithm2[0]);
            String intern4 = typeAndAlgorithm2[1].intern();
            Provider.ServiceKey serviceKey3 = new Provider.ServiceKey(engineName3, s2, true);
            Provider.Service service3 = (Provider.Service)this.legacyMap.get(serviceKey3);
            if (service3 == null) {
                service3 = new Provider.Service(this);
                service3.type = engineName3;
                service3.algorithm = s2;
                this.legacyMap.put(serviceKey3, service3);
            }
            this.legacyMap.put(new Provider.ServiceKey(engineName3, intern4, true), service3);
            Provider.Service.access$700(service3, intern4);
        }
        else {
            String[] typeAndAlgorithm = getTypeAndAlgorithm(s);
            if (typeAndAlgorithm == null) {
                return;
            }
            int index = typeAndAlgorithm[1].indexOf(32);
            if (index == -1) {
                String engineName2 = getEngineName(typeAndAlgorithm[0]);
                String intern3 = typeAndAlgorithm[1].intern();
                Provider.ServiceKey serviceKey2 = new Provider.ServiceKey(engineName2, intern3, true);
                Provider.Service service2 = (Provider.Service)this.legacyMap.get(serviceKey2);
                if (service2 == null) {
                    service2 = new Provider.Service(this);
                    service2.type = engineName2;
                    service2.algorithm = intern3;
                    this.legacyMap.put(serviceKey2, service2);
                }
                service2.className = s2;
                if (engineName2.equals("SecureRandom")) {
                    this.updateSecureRandomEntries(true, service2.algorithm);
                }
            }
            else {
                String engineName = getEngineName(typeAndAlgorithm[0]);
                String s3 = typeAndAlgorithm[1];
                String intern = s3.substring(0, index).intern();
                String s4;
                for (s4 = s3.substring(index + 1); s4.startsWith(" "); s4 = s4.substring(1)) {}
                String intern2 = s4.intern();
                Provider.ServiceKey serviceKey = new Provider.ServiceKey(engineName, intern, true);
                Provider.Service service = (Provider.Service)this.legacyMap.get(serviceKey);
                if (service == null) {
                    service = new Provider.Service(this);
                    service.type = engineName;
                    service.algorithm = intern;
                    this.legacyMap.put(serviceKey, service);
                }
                service.addAttribute(intern2, s2);
            }
        }
    }
    
    public Provider.Service getService(String s, String s2) {
        this.checkInitialized();
        Provider.ServiceKey previousKey = Provider.previousKey;
        if (!previousKey.matches(s, s2)) {
            previousKey = (Provider.previousKey = new Provider.ServiceKey(s, s2, false));
        }
        if (!this.serviceMap.isEmpty()) {
            Provider.Service service = (Provider.Service)this.serviceMap.get(previousKey);
            if (service != null) {
                return service;
            }
        }
        synchronized (this) {
            this.ensureLegacyParsed();
            if (this.legacyMap != null && !this.legacyMap.isEmpty()) {
                return (Provider.Service)this.legacyMap.get(previousKey);
            }
        }
        return null;
    }
    
    public synchronized Set<Provider.Service> getServices() {
        this.checkInitialized();
        if (this.legacyChanged || this.servicesChanged) {
            this.serviceSet = null;
        }
        if (this.serviceSet == null) {
            this.ensureLegacyParsed();
            LinkedHashSet<Object> s = new LinkedHashSet();
            if (!this.serviceMap.isEmpty()) {
                s.addAll(this.serviceMap.values());
            }
            if (this.legacyMap != null && !this.legacyMap.isEmpty()) {
                s.addAll(this.legacyMap.values());
            }
            this.serviceSet = Collections.unmodifiableSet(s);
            this.servicesChanged = false;
        }
        return this.serviceSet;
    }
    
    protected void putService(Provider.Service obj) {
        this.check("putProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println(this.name + ".putService(): " + obj);
        }
        if (obj == null) {
            throw new NullPointerException();
        }
        if (obj.getProvider() != this) {
            throw new IllegalArgumentException("service.getProvider() must match this Provider object");
        }
        String type = obj.getType();
        Provider.ServiceKey serviceKey = new Provider.ServiceKey(type, obj.getAlgorithm(), true);
        this.implRemoveService((Provider.Service)this.serviceMap.get(serviceKey));
        this.serviceMap.put(serviceKey, obj);
        Iterator<Object> iterator = Provider.Service.access$800(obj).iterator();
        while (iterator.hasNext()) {
            this.serviceMap.put(new Provider.ServiceKey(type, (String)iterator.next(), true), obj);
        }
        this.servicesChanged = true;
        synchronized (this) {
            this.putPropertyStrings(obj);
            if (type.equals("SecureRandom")) {
                this.updateSecureRandomEntries(true, obj.algorithm);
            }
        }
        return;
    }
    
    private void updateSecureRandomEntries(boolean b, String s) {
        Objects.requireNonNull(s);
        if (b) {
            if (this.prngAlgos == null) {
                this.prngAlgos = new LinkedHashSet();
            }
            this.prngAlgos.add(s);
        }
        else {
            this.prngAlgos.remove(s);
        }
        if (Provider.debug != null) {
            Provider.debug.println((b ? "Add" : "Remove") + " SecureRandom algo " + s);
        }
    }
    
    synchronized Provider.Service getDefaultSecureRandomService() {
        this.checkInitialized();
        if (this.legacyChanged) {
            this.prngAlgos = null;
            this.ensureLegacyParsed();
        }
        if (this.prngAlgos != null && !this.prngAlgos.isEmpty()) {
            return this.getService("SecureRandom", (String)this.prngAlgos.iterator().next());
        }
        return null;
    }
    
    private void putPropertyStrings(Provider.Service service) {
        String type = service.getType();
        String algorithm = service.getAlgorithm();
        super.put(type + "." + algorithm, service.getClassName());
        Iterator<Object> iterator = Provider.Service.access$800(service).iterator();
        while (iterator.hasNext()) {
            super.put("Alg.Alias." + type + "." + (String)iterator.next(), algorithm);
        }
        for (Map.Entry entry : service.attributes.entrySet()) {
            super.put(type + "." + algorithm + " " + entry.getKey(), entry.getValue());
        }
    }
    
    private void removePropertyStrings(Provider.Service service) {
        String type = service.getType();
        String algorithm = service.getAlgorithm();
        super.remove(type + "." + algorithm);
        Iterator<Object> iterator = Provider.Service.access$800(service).iterator();
        while (iterator.hasNext()) {
            super.remove("Alg.Alias." + type + "." + (String)iterator.next());
        }
        Iterator<Object> iterator2 = service.attributes.entrySet().iterator();
        while (iterator2.hasNext()) {
            super.remove(type + "." + algorithm + " " + ((Map.Entry)iterator2.next()).getKey());
        }
    }
    
    protected void removeService(Provider.Service obj) {
        this.check("removeProviderProperty." + this.name);
        if (Provider.debug != null) {
            Provider.debug.println(this.name + ".removeService(): " + obj);
        }
        if (obj == null) {
            throw new NullPointerException();
        }
        this.implRemoveService(obj);
        return;
    }
    
    private void implRemoveService(Provider.Service service) {
        if (service == null || this.serviceMap.isEmpty()) {
            return;
        }
        String type = service.getType();
        Provider.ServiceKey serviceKey = new Provider.ServiceKey(type, service.getAlgorithm(), false);
        if (service != (Provider.Service)this.serviceMap.get(serviceKey)) {
            return;
        }
        this.servicesChanged = true;
        this.serviceMap.remove(serviceKey);
        Iterator<Object> iterator = Provider.Service.access$800(service).iterator();
        while (iterator.hasNext()) {
            this.serviceMap.remove(new Provider.ServiceKey(type, (String)iterator.next(), false));
        }
        synchronized (this) {
            this.removePropertyStrings(service);
            if (type.equals("SecureRandom")) {
                this.updateSecureRandomEntries(false, service.algorithm);
            }
        }
        return;
    }
    
    private static void addEngine(String s, boolean b, String s2) {
        Provider.EngineDescription engineDescription = new EngineDescription(s, b, s2);
        Provider.knownEngines.put(s.toLowerCase(Locale.ENGLISH), engineDescription);
        Provider.knownEngines.put(s, engineDescription);
    }
    
    private static String getEngineName(String s) {
        Provider.EngineDescription engineDescription = (Provider.EngineDescription)Provider.knownEngines.get(s);
        if (engineDescription == null) {
            engineDescription = (Provider.EngineDescription)Provider.knownEngines.get(s.toLowerCase(Locale.ENGLISH));
        }
        return (engineDescription == null) ? s : engineDescription.name;
    }
    
    static String access$1000(String s) {
        return getEngineName(s);
    }
    
    static {
        debug = Debug.getInstance("provider", "Provider");
        ALIAS_LENGTH = "Alg.Alias.".length();
        Provider.previousKey = new Provider.ServiceKey("", "", false);
        knownEngines = new HashMap();
        addEngine("AlgorithmParameterGenerator", false, null);
        addEngine("AlgorithmParameters", false, null);
        addEngine("KeyFactory", false, null);
        addEngine("KeyPairGenerator", false, null);
        addEngine("KeyStore", false, null);
        addEngine("MessageDigest", false, null);
        addEngine("SecureRandom", false, null);
        addEngine("Signature", true, null);
        addEngine("CertificateFactory", false, null);
        addEngine("CertPathBuilder", false, null);
        addEngine("CertPathValidator", false, null);
        addEngine("CertStore", false, "java.security.cert.CertStoreParameters");
        addEngine("Cipher", true, null);
        addEngine("ExemptionMechanism", false, null);
        addEngine("Mac", true, null);
        addEngine("KeyAgreement", true, null);
        addEngine("KeyGenerator", false, null);
        addEngine("SecretKeyFactory", false, null);
        addEngine("KeyManagerFactory", false, null);
        addEngine("SSLContext", false, null);
        addEngine("TrustManagerFactory", false, null);
        addEngine("GssApiMechanism", false, null);
        addEngine("SaslClientFactory", false, null);
        addEngine("SaslServerFactory", false, null);
        addEngine("Policy", false, "java.security.Policy$Parameters");
        addEngine("Configuration", false, "javax.security.auth.login.Configuration$Parameters");
        addEngine("XMLSignatureFactory", false, null);
        addEngine("KeyInfoFactory", false, null);
        addEngine("TransformService", false, null);
        addEngine("TerminalFactory", false, "java.lang.Object");
    }
    
    private static class ServiceKey
    {
        private final String type;
        private final String algorithm;
        private final String originalAlgorithm;
        
        private ServiceKey(String type, String upperCase, boolean b) {
            this.type = type;
            this.originalAlgorithm = upperCase;
            upperCase = upperCase.toUpperCase(Locale.ENGLISH);
            this.algorithm = (b ? upperCase.intern() : upperCase);
        }
        
        @Override
        public int hashCode() {
            return Objects.hash(new Object[] { this.type, this.algorithm });
        }
        
        @Override
        public boolean equals(Object o) {
            if (this == o) {
                return true;
            }
            if (!(o instanceof Provider.ServiceKey)) {
                return false;
            }
            Provider.ServiceKey serviceKey = (Provider.ServiceKey)o;
            return this.type.equals(serviceKey.type) && this.algorithm.equals(serviceKey.algorithm);
        }
        
        boolean matches(String s, String s2) {
            return this.type == s && this.originalAlgorithm == s2;
        }
        
        @Override
        public String toString() {
            return this.type + ":" + this.algorithm;
        }
    }
    
    private static class UString
    {
        final String string;
        final String lowerString;
        
        UString(String string) {
            this.string = string;
            this.lowerString = string.toLowerCase(Locale.ENGLISH);
        }
        
        @Override
        public int hashCode() {
            return this.lowerString.hashCode();
        }
        
        @Override
        public boolean equals(Object o) {
            return this == o || (o instanceof Provider.UString && this.lowerString.equals(((Provider.UString)o).lowerString));
        }
        
        @Override
        public String toString() {
            return this.string;
        }
    }
    
    private static class EngineDescription
    {
        final String name;
        final boolean supportsParameter;
        final String constructorParameterClassName;
        private volatile Class<?> constructorParameterClass;
        
        EngineDescription(String name, boolean supportsParameter, String constructorParameterClassName) {
            this.name = name;
            this.supportsParameter = supportsParameter;
            this.constructorParameterClassName = constructorParameterClassName;
        }
        
        Class<?> getConstructorParameterClass() throws ClassNotFoundException {
            Class<Object> constructorParameterClass = this.constructorParameterClass;
            if (constructorParameterClass == null) {
                constructorParameterClass = Class.forName(this.constructorParameterClassName);
                this.constructorParameterClass = constructorParameterClass;
                return constructorParameterClass;
            }
            return constructorParameterClass;
        }
    }
    
    public static class Service
    {
        private String type;
        private String algorithm;
        private String className;
        private final Provider provider;
        private List<String> aliases;
        private Map<Provider.UString, String> attributes;
        private volatile Reference<Class<?>> classRef;
        private volatile Boolean hasKeyAttributes;
        private String[] supportedFormats;
        private Class[] supportedClasses;
        private boolean registered;
        private static final Class<?>[] CLASS0;
        
        private Service(Provider provider) {
            this.provider = provider;
            this.aliases = Collections.emptyList();
            this.attributes = Collections.emptyMap();
        }
        
        private boolean isValid() {
            return this.type != null && this.algorithm != null && this.className != null;
        }
        
        private void addAlias(String s) {
            if (this.aliases.isEmpty()) {
                this.aliases = new ArrayList(2);
            }
            this.aliases.add(s);
        }
        
        void addAttribute(String s, String s2) {
            if (this.attributes.isEmpty()) {
                this.attributes = new HashMap(8);
            }
            this.attributes.put(new UString(s), s2);
        }
        
        public Service(Provider provider, String s, String algorithm, String className, List<String> c, Map<String, String> map) {
            if (provider == null || s == null || algorithm == null || className == null) {
                throw new NullPointerException();
            }
            this.provider = provider;
            this.type = Provider.access$1000(s);
            this.algorithm = algorithm;
            this.className = className;
            if (c == null) {
                this.aliases = Collections.emptyList();
            }
            else {
                this.aliases = new ArrayList(c);
            }
            if (map == null) {
                this.attributes = Collections.emptyMap();
            }
            else {
                this.attributes = new HashMap();
                for (Map.Entry entry : map.entrySet()) {
                    this.attributes.put(new Provider.UString((String)entry.getKey()), entry.getValue());
                }
            }
            return;
        }
        
        public final String getType() {
            return this.type;
        }
        
        public final String getAlgorithm() {
            return this.algorithm;
        }
        
        public final Provider getProvider() {
            return this.provider;
        }
        
        public final String getClassName() {
            return this.className;
        }
        
        private final List<String> getAliases() {
            return this.aliases;
        }
        
        public final String getAttribute(String s) {
            if (s == null) {
                throw new NullPointerException();
            }
            return (String)this.attributes.get(new UString(s));
        }
        
        public Object newInstance(Object o) throws NoSuchAlgorithmException {
            if (!this.registered) {
                if (this.provider.getService(this.type, this.algorithm) != this) {
                    throw new NoSuchAlgorithmException("Service not registered with Provider " + this.provider.getName() + ": " + this);
                }
                this.registered = true;
            }
            try {
                Provider.EngineDescription engineDescription = (Provider.EngineDescription)Provider.knownEngines.get(this.type);
                if (engineDescription == null) {
                    return this.newInstanceGeneric(o);
                }
                if (engineDescription.constructorParameterClassName == null) {
                    if (o != null) {
                        throw new InvalidParameterException("constructorParameter not used with " + this.type + " engines");
                    }
                    return this.getImplClass().getConstructor(new Class[0]).newInstance(new Object[0]);
                }
                else {
                    Class<Object> constructorParameterClass = engineDescription.getConstructorParameterClass();
                    if (o != null && !constructorParameterClass.isAssignableFrom(o.getClass())) {
                        throw new InvalidParameterException("constructorParameter must be instanceof " + engineDescription.constructorParameterClassName.replace('$', '.') + " for engine type " + this.type);
                    }
                    return this.getImplClass().getConstructor(new Class[] { constructorParameterClass }).newInstance(new Object[] { o });
                }
            }
            catch (NoSuchAlgorithmException ex) {
                throw ex;
            }
            catch (InvocationTargetException ex2) {
                throw new NoSuchAlgorithmException("Error constructing implementation (algorithm: " + this.algorithm + ", provider: " + this.provider.getName() + ", class: " + this.className + ")", ex2.getCause());
            }
            catch (Exception ex3) {
                throw new NoSuchAlgorithmException("Error constructing implementation (algorithm: " + this.algorithm + ", provider: " + this.provider.getName() + ", class: " + this.className + ")", ex3);
            }
        }
        
        private Class<?> getImplClass() throws NoSuchAlgorithmException {
            try {
                Reference<Object> classRef = this.classRef;
                Class<Object> referent = (classRef == null) ? null : ((Class)classRef.get());
                if (referent == null) {
                    ClassLoader classLoader = this.provider.getClass().getClassLoader();
                    if (classLoader == null) {
                        referent = Class.forName(this.className);
                    }
                    else {
                        referent = classLoader.loadClass(this.className);
                    }
                    if (!Modifier.isPublic(referent.getModifiers())) {
                        throw new NoSuchAlgorithmException("class configured for " + this.type + " (provider: " + this.provider.getName() + ") is not public.");
                    }
                    this.classRef = new WeakReference(referent);
                }
                return referent;
            }
            catch (ClassNotFoundException ex) {
                throw new NoSuchAlgorithmException("class configured for " + this.type + " (provider: " + this.provider.getName() + ") cannot be found.", ex);
            }
        }
        
        private Object newInstanceGeneric(Object o) throws Exception {
            Class<Object> implClass = this.getImplClass();
            if (o == null) {
                try {
                    return implClass.getConstructor(new Class[0]).newInstance(new Object[0]);
                }
                catch (NoSuchMethodException ex) {
                    throw new NoSuchAlgorithmException("No public no-arg constructor found in class " + this.className);
                }
            }
            else {
                Class<Object> class1 = o.getClass();
                Constructor[] constructors;
                for (Constructor constructor : constructors = implClass.getConstructors()) {
                    Class[] parameterTypes = constructor.getParameterTypes();
                    if (parameterTypes.length == 1 && parameterTypes[0].isAssignableFrom(class1)) {
                        return constructor.newInstance(new Object[] { o });
                    }
                }
                throw new NoSuchAlgorithmException("No public constructor matching " + class1.getName() + " found in class " + this.className);
            }
        }
        
        public boolean supportsParameter(Object o) {
            Provider.EngineDescription engineDescription = (Provider.EngineDescription)Provider.knownEngines.get(this.type);
            if (engineDescription == null) {
                return true;
            }
            if (!engineDescription.supportsParameter) {
                throw new InvalidParameterException("supportsParameter() not used with " + this.type + " engines");
            }
            if (o != null && !(o instanceof Key)) {
                throw new InvalidParameterException("Parameter must be instanceof Key for engine " + this.type);
            }
            if (!this.hasKeyAttributes()) {
                return true;
            }
            if (o == null) {
                return false;
            }
            Key key = (Key)o;
            return this.supportsKeyFormat(key) || this.supportsKeyClass(key);
        }
        
        private boolean hasKeyAttributes() {
            Boolean hasKeyAttributes = this.hasKeyAttributes;
            if (hasKeyAttributes == null) {
                synchronized (this) {
                    String attribute = this.getAttribute("SupportedKeyFormats");
                    if (attribute != null) {
                        this.supportedFormats = attribute.split("\\|");
                    }
                    String attribute2 = this.getAttribute("SupportedKeyClasses");
                    if (attribute2 != null) {
                        String[] split = attribute2.split("\\|");
                        ArrayList list = new ArrayList(split.length);
                        for (int length = split.length, i = 0; i < length; ++i) {
                            Class<Object> keyClass = this.getKeyClass(split[i]);
                            if (keyClass != null) {
                                list.add(keyClass);
                            }
                        }
                        this.supportedClasses = (Class[])list.toArray(Provider.Service.CLASS0);
                    }
                    hasKeyAttributes = Boolean.valueOf(this.supportedFormats != null || this.supportedClasses != null);
                    this.hasKeyAttributes = hasKeyAttributes;
                }
            }
            return (boolean)hasKeyAttributes;
        }
        
        private Class<?> getKeyClass(String s) {
            try {
                return Class.forName(s);
            }
            catch (ClassNotFoundException ex) {
                try {
                    ClassLoader classLoader = this.provider.getClass().getClassLoader();
                    if (classLoader != null) {
                        return classLoader.loadClass(s);
                    }
                }
                catch (ClassNotFoundException ex2) {}
                return null;
            }
        }
        
        private boolean supportsKeyFormat(Key key) {
            if (this.supportedFormats == null) {
                return false;
            }
            String format = key.getFormat();
            if (format == null) {
                return false;
            }
            String[] supportedFormats = this.supportedFormats;
            int length = supportedFormats.length;
            int i = 0;
            while (i < length) {
                if (supportedFormats[i].equals(format)) {
                    return true;
                }
                ++i;
            }
            return false;
        }
        
        private boolean supportsKeyClass(Key key) {
            if (this.supportedClasses == null) {
                return false;
            }
            Class<Object> class1 = key.getClass();
            Class[] supportedClasses = this.supportedClasses;
            int length = supportedClasses.length;
            int i = 0;
            while (i < length) {
                if (supportedClasses[i].isAssignableFrom(class1)) {
                    return true;
                }
                ++i;
            }
            return false;
        }
        
        @Override
        public String toString() {
            return this.provider.getName() + ": " + this.type + "." + this.algorithm + " -> " + this.className + (this.aliases.isEmpty() ? "" : ("\r\n  aliases: " + this.aliases.toString())) + (this.attributes.isEmpty() ? "" : ("\r\n  attributes: " + this.attributes.toString())) + "\r\n";
        }
        
        static boolean access$000(Provider.Service service) {
            return service.isValid();
        }
        
        static void access$700(Provider.Service service, String s) {
            service.addAlias(s);
        }
        
        static List access$800(Provider.Service service) {
            return service.getAliases();
        }
        
        static {
            CLASS0 = new Class[0];
        }
    }
}
